;====================================================================
.model small, c
.386
.stack
;====================================================================
include support.asm
;====================================================================

LOADOVERLAY STRUCT
    loStartSegment      dw ?
    loRelocationFactor  dw ?
LOADOVERLAY ENDS

;====================================================================
.data

    ;--------------------------------------------------------------
    ; MASM treats these as FAR pointers because of the DWORD size.
    ; The initialized values are the offsets into the jump table,
    ; so the calls are actually directed to the jump table, and
    ; the jumps in the table route them to the procedures, without
    ; disturbing any arguments on the stack.
    ;--------------------------------------------------------------

    fp_proc0      dd 0
    fp_proc1      dd 6
    fp_proc2      dd 12

    overlay_name  db "overlay.exe",0

    lo            LOADOVERLAY <>

.code
.startup
;====================================================================

    ;---------------------------------------------------------
    ; Get and display the segment address of our PSP, back up
    ; to the preceding memory control block and display the
    ; signature, owner, and size fields. The signature field
    ; will be 4Dh, or 5Ah for the last block. The size field
    ; contains the size of the memory block in paragraphs,
    ; here converted and displayed in bytes.
    ;---------------------------------------------------------

    mov ah, 51h
    int 21h
    push bx
    push bx
    print "PSP seg       = ",hexword$(bx),"h",NL
    sub bx, 1
    mov es, bx
    xor bx, bx
    print "MCB signature = ",hexbyte$(es:[bx]),"h",NL
    print "MCB owner     = ",hexword$(es:[bx+1]),"h",NL
    movzx eax, WORD PTR es:[bx+3]
    shl eax,4
    print "MCB size      = ",dword$(eax),NL,NL

    ;-------------------------------------------
    ; Display the values of CS, DS, SS. and SP.
    ;-------------------------------------------

    print "CS = ",word$(cs),"h",NL
    print "DS = ",word$(ds),"h",NL
    print "SS = ",word$(ss),"h",NL
    print "SP = ",word$(sp),"h",NL,NL

    ;-----------------------------------------------------------
    ; Calculate the actual memory footprint, in paragraphs, as
    ; (PSP seg) - (SS + (SP SHR 4) + 1), and then resize our
    ; memory block to this size.
    ;-----------------------------------------------------------

    pop bx                  ; recover PSP seg
    mov es, bx
    mov dx, ss
    mov ax, sp
    shr ax, 4
    inc ax                  ; adjust for bits shifted out
    add ax, dx
    sub ax, bx
    mov bx, ax
    mov ah, 4ah             ; Set Memory Block Size
    int 21h
    jnc @F
    print "resize error",NL
    call waitkey
    .exit
  @@:

    ;-------------------------------------------------------------
    ; Display the new signature and the size of the memory block.
    ;-------------------------------------------------------------

    pop bx                  ; recover PSP seg
    sub bx, 1
    mov es, bx
    xor bx, bx
    print "MCB signature = ",hexbyte$(es:[bx]),"h",NL
    movzx eax, WORD PTR es:[bx+3]
    shl eax,4
    print "MCB size      = ",dword$(eax),NL,NL

    ;--------------------------------
    ; Allocate 16KB for the overlay.
    ;--------------------------------

    mov bx, 1024
    mov ah, 48h
    int 21h
    jnc @F
    print "allocation error",NL
    call waitkey
    .exit
  @@:

    ;-----------------------------------------------------------
    ; Store the segment address of the allocated memory to our
    ; procedure pointers and to both members of the LOADOVERLAY
    ; structure.
    ;-----------------------------------------------------------

    mov WORD PTR fp_proc0+2, ax
    mov WORD PTR fp_proc1+2, ax
    mov WORD PTR fp_proc2+2, ax

    mov lo.loStartSegment, ax
    mov lo.loRelocationFactor, ax

    ;-------------------
    ; Load the overlay.
    ;-------------------

    mov dx, OFFSET overlay_name
    push SEG lo
    pop es
    mov bx, OFFSET lo
    mov ax, 4b03h
    int 21h
    jnc @F
    print "load overlay error",NL
    call waitkey
    .exit
  @@:

    ;----------------------
    ; Call the procedures.
    ;----------------------

    push DWORD PTR 1234
    call fp_proc0
    print dword$(eax),NL
    push DWORD PTR 1234
    call fp_proc1
    print dword$(eax),NL
    push DWORD PTR 1234
    call fp_proc2
    print dword$(eax),NL

    ;----------------------------
    ; Free the allocated memory.
    ;----------------------------

    mov ax, lo.loStartSegment
    mov es, ax
    mov ah, 49h
    int 21h

    call waitkey
    .exit

end
